# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import bisect
from collections import defaultdict
import json
from tekton.gae.middleware import Middleware


class _ParamExtractor(object):
    def __init__(self):
        self.indexed_values = defaultdict(list)
        self.indexed_indexes = defaultdict(list)

    def _extract_values(self, handler, param, default_value=""):
        values = handler.request.get_all(param)
        if param.endswith('[]'):
            return param[:-2], values if values else []
        elif param.endswith(']') and '[' in param:
            try:
                param_name, param_idx = param[:-1].split('[')
                param_idx = int(param_idx)
                sorted_list = self.indexed_values[param_name]
                sorted_indexes = self.indexed_indexes[param_name]
                idx = bisect.bisect_right(sorted_indexes, param_idx)
                sorted_list.insert(idx, values)
                sorted_indexes.insert(idx, param_idx)
                return param_name, sorted_list
            except:
                pass

        if not values:
            return param, default_value
        if len(values) == 1:
            return param, values[0]
        return param, values


class RequestParamsMiddleware(Middleware):
    def set_up(self):
        json_header = r'application/json, text/plain, */*'
        header_value = getattr(self.handler.request, 'accept', None)
        header_value = getattr(header_value, 'header_value', None)
        # AngularJS Call
        if header_value == json_header and self.handler.request.body:
            self.request_args.update(json.loads(self.handler.request.body))
        else:
            extractor = _ParamExtractor()
            self.request_args.update(
                dict(extractor._extract_values(self.handler, a) for a in self.handler.request.arguments()))


# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals


class Middleware(object):
    def __init__(self, handler, dependencies, request_args):
        self.dependencies = dependencies
        self.handler = handler
        self.request_args = request_args

    def set_up(self):
        """
        Must return True to stop next middleware execution
        """
        pass

    def tear_down(self):
        pass

    def handle_error(self, e):
        pass


def execute(middleware_classes, handler):
    """
    This version of execute uses Object Orientation to keep stack trace clean
    Middlewares must be Middleware classes to be executed
    """
    dependecies = {}
    request_args = {}
    executed_middlewares = []
    exception = None
    for mid_class in middleware_classes:
        mid_obj = mid_class(handler, dependecies, request_args)
        executed_middlewares.append(mid_obj)
        try:
            should_stop_next_middleware = mid_obj.set_up()
            if should_stop_next_middleware:
                break
        except Exception, e:
            exception = e
            break

    for mid in reversed(executed_middlewares):
        if exception:
            mid.handle_error(exception)
        else:
            mid.tear_down()





# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import json
from tekton.gae.middleware import Middleware
from tekton.gae.middleware.response import ResponseBase


class JsonResponse(ResponseBase):
    def __init__(self, context, secure_prefix=")]}',\n"):
        super(JsonResponse, self).__init__(context)
        self.secure_prefix = secure_prefix

    def to_json(self):
        return self.secure_prefix + json.dumps(self.context)


class JsonUnsecureResponse(JsonResponse):
    def __init__(self, context, secure_prefix=''):
        super(JsonUnsecureResponse, self).__init__(context, secure_prefix)


class JsonResponseMiddleware(Middleware):
    def set_up(self):
        fcn_response = self.dependencies['_fcn_response']
        if isinstance(fcn_response, JsonResponse):
            resp = self.handler.response
            resp.headers[str('Content-Type')] = str('application/json')
            resp.write(fcn_response.to_json())
            return True  # after response, there is no need to look for more middlewares






# -*- coding]= utf-8 -*-
from __future__ import absolute_import, unicode_literals
from tekton.gae.middleware import Middleware


class Webapp2Dependencies(Middleware):
    def set_up(self):
        self.dependencies["_req"] = self.handler.request
        self.dependencies["_resp"] = self.handler.response
        self.dependencies["_handler"] = self.handler
        self.dependencies["_dependencies"] = self.dependencies


# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import json
import logging
import traceback
import time

from google.appengine.api import app_identity, mail, capabilities
from google.appengine.runtime import DeadlineExceededError
from tekton.gae.middleware import Middleware

from tekton.router import PathNotFound


def get_apis_statuses(e):
    if not isinstance(e, DeadlineExceededError):
        return {}
    t1 = time.time()
    statuses = {
        'blobstore': capabilities.CapabilitySet('blobstore').is_enabled(),
        'datastore_v3': capabilities.CapabilitySet('datastore_v3').is_enabled(),
        'datastore_v3_write': capabilities.CapabilitySet('datastore_v3', ['write']).is_enabled(),
        'images': capabilities.CapabilitySet('images').is_enabled(),
        'mail': capabilities.CapabilitySet('mail').is_enabled(),
        'memcache': capabilities.CapabilitySet('memcache').is_enabled(),
        'taskqueue': capabilities.CapabilitySet('taskqueue').is_enabled(),
        'urlfetch': capabilities.CapabilitySet('urlfetch').is_enabled(),
    }
    t2 = time.time()
    statuses['time'] = t2 - t1
    return statuses


def send_error_to_admins(settings, exception, handler, render, template):
    tb = traceback.format_exc()
    errmsg = exception.message

    logging.error(errmsg)
    logging.error(tb)
    handler.response.write(render(template))
    appid = app_identity.get_application_id()

    subject = 'ERROR in %s: [%s] %s' % (appid, handler.request.path, errmsg)
    body = """
------------- request ------------
%s
----------------------------------

------------- GET params ---------
%s
----------------------------------

----------- POST params ----------
%s
----------------------------------

----------- traceback ------------
%s
----------------------------------
""" % (handler.request, handler.request.GET, handler.request.POST, tb)
    body += 'API statuses = ' + json.dumps(get_apis_statuses(exception), indent=4)
    mail.send_mail_to_admins(sender=settings.SENDER_EMAIL,
                             subject=subject,
                             body=body)


class EmailMiddleware(Middleware):
    def handle_error(self, exception):
        import settings  # workaround. See https://github.com/renzon/zenwarch/issues/3

        if isinstance(exception, PathNotFound):
            self.handler.response.set_status(404)
            send_error_to_admins(settings, exception, self.handler, self.dependencies['_render'],
                                 settings.TEMPLATE_404_ERROR)
        else:
            self.handler.response.set_status(400)
            send_error_to_admins(settings, exception, self.handler, self.dependencies['_render'],
                                 settings.TEMPLATE_400_ERROR)

